function [ mImageApprox, vCoefficients, iD1, iV1, iD2, iV2, k, i1D ] =  ApproxRDCTDB(mImage, mDictionary, tolerance, blockWidth)
% ApproxRDCTDB Approximates a <latex>2<latex>D signal using the RDCTDB dictionary and the OMP algorithm.
%
% Implementation of OMP in \(2\)D using \(2\) separable \(1\)D dictionaries (for rows and columns)
% The dictionaries used here are \(D_{a}\) and \(D_{b}\) where:
%   \(D_{a}\) - Normalized redundancy \(2\) discrete cosine dictionary.
%   \(D_{b}\) - Dirac basis.
% This implementation uses the fast DCT to calculate the inner products with the residual as
% in equations (3)-(5) of [1].
%
% Usage:    [ mImageApprox, vCoefficients, iD1, iV1, iD2, iV2, k, i1D ] =  ApproxRDCTDB( ...
%               mImage, mDictionary, tolerance, blockWidth );
%
% Inputs:
%   mImage          Matrix representing a block of the image.
%   mDictionary     Dictionary of normalized atoms: \([D_{a}, D_{b}]\otimes[D_{a}, D_{b}]\);
%   tolerance       Stop when norm of the residual \(<\) tolerance. 
%   blockWidth      Width of the square matrix mImage. 
%
% Outputs:
%   mImageApprox    Approximation of mImage.
%   vCoefficients   Coefficients of the decomposition.
%   iD1             Indices of the dictionary to which the selected atom belongs (w.r.t 
%                   columns) iD1 = 1 indicates that the selected atom belongs to dictionary
%                   \(D_{a}\).
%   iV1             Indices of the selected atom in the dictionary iD1.
%   iD2             Indices of the dictionary to which the selected atom belongs (w.r.t.
%                   rows).
%   iV2             Indices of the selected atom in the dictionary iD2.
%   k               Number of selected atoms.
%   i1D             Indices of the selected atoms in mDictionary.
%
% The implementation of OMP method is based on  Gram-Schmidt orthonormalization and adaptive
% birothogonalization, as proposed in Ref 2 below.
%
% References:
% 
% 1-Y.C. Pati, R. Rezaiifar, and P.S. Krishnaprasad, "Orthogonal matching pursuits:
% recursive function approximation with applications to wavelet decomposition", in Proceedings of
% the 27th Asilomar Conference on Signals, Systems and Computers, 1993.
% 
% 2-L. Rebollo-Neira and D. Lowe, "Optimized Orthogonal Matching Pursuit Approach",
% IEEE Signal Processing Letters, Vol(9,4), 137-140, 2002.


    functionName = mfilename;

    if nargin < 2

        error('%s: Not enough input arguments!',functionName);

    end

    if nargin < 3

        if size(mImage,1) ~= size(mImage,2)

            error('%s: The image blocks are not square!',functionName)

        end

        blockWidth = size(mImage,1);

    end   

    mResidule = mImage;
    vImage = mImage(:)';

    % number of reorthogonalizations
    ZMAX = 2;
    % threshold for coefficients
    tol2 = 1e-10;

    % Initialize matrices
    iD1 = zeros(1,blockWidth^2); 
    iD2 = zeros(1,blockWidth^2); 
    iV1 = zeros(1,blockWidth^2); 
    iV2 = zeros(1,blockWidth^2);
    i1D = zeros(1,blockWidth^2);

    nDCTAtoms = (2*blockWidth) - 1;
    SQRT2 = sqrt(2);

    %=================================
    % Main algorithm: at kth iteration
    %=================================
    maxIterations = blockWidth^2;
    for k = 1:maxIterations

        [ iD1(k), iV1(k), iD2(k), iV2(k), maxCn ] = ChooseAtom( mResidule, nDCTAtoms, SQRT2 );
        

        
        if maxCn < tol2

            if k == 1
                i1D(k) = 1;
            end
            k=k-1;           
            break;

        end    
        
        ind = KroneckerIndex(iD2(k),iD1(k),iV2(k),iV1(k),[ blockWidth*2-1, blockWidth]);
        i1D(k) = ind; 
        vAtomK = mDictionary(:,ind);
     
        if k > 1

            %Q(:,k) is the orthogonalization of D(:,k)  w.r.t Q(:,1),..., Q(:,k-1) 
            for p=1:k-1

                Q(:,k) = vAtomK - (Q(:,p)'*vAtomK)*Q(:,p); %orthogonalization

            end

            %re-orthogonalization of Q(:,k)  w.r.t Q(:,1),..., Q(:,k-1)
            for zi=1:ZMAX

                for p=1:k-1

                    Q(:,k) = Q(:,k)-(Q(:,p)'*Q(:,k))*Q(:,p); %re-orthogonalization

                end

            end

        else

           Q(:,k) = vAtomK; 

        end

        nork=norm(Q(:,k));
        Q(:,k) = Q(:,k)/nork;

        if k > 1

            mBeta = mBeta - Q(:,k)*(vAtomK'*mBeta)/nork;

        end

        mBeta(:,k) = Q(:,k)/nork;

        temp = mResidule(:)' - vImage*Q(:,k)*Q(:,k)';
        mResidule(:) = temp;  

        if (tolerance~= 0) && (norm(temp) < tolerance)

            break;

        end

    end

    % If no atoms chosen return a zero coefficient at position 1.
    if k == 0
        k = 1;
        Q = ones(blockWidth^2,1);
        mBeta = ones(blockWidth^2,1);
        vCoefficients = 0;
    else
        % Reduce the size of the matrices to the number of atoms chosen
        Q = Q(:,1:k);
        mBeta = mBeta(:,1:k);
        vCoefficients = (vImage*mBeta)';
    end

    % Compare with original minus residule
    mImageApprox = mImage - mResidule;

end

function [ i1, i2, maxCn ] = getMaxIndex( mCn )

    maxCn = 0;
    i1 = 1;
    i2 = 1;

    for i = 1:size(mCn,1)

        for j = 1:size(mCn,2)

            if abs(mCn(i,j)) > maxCn

                maxCn = abs(mCn(i,j));
                i1 = i;
                i2 = j;

            end

        end

    end
end

function [ iD1, iV1, iD2, iV2, maxCn ] = ChooseAtom( mResidule, nDCTAtoms, SQRT2 )
        
    iD1 = 1;
    iD2 = 1;
    mCDCT2 = dct2(mResidule,nDCTAtoms+1,nDCTAtoms+1)*2;
    [ iV1, iV2, maxCn ] = getMaxIndex( mCDCT2(1:nDCTAtoms,1:nDCTAtoms) );
    mCDCTDB = dct(mResidule,nDCTAtoms+1);
    [ iTemp1, iTemp2, maxTemp ] = getMaxIndex( mCDCTDB(1:nDCTAtoms,:) );    
    if maxTemp*SQRT2 > maxCn
        
        maxCn = maxTemp*SQRT2;
        iD1 = 1;
        iD2 = 2;
        iV1 = iTemp1;
        iV2 = iTemp2;
        
    end
    
    mCDBDCT = dct(mResidule',nDCTAtoms+1);
    [ iTemp1, iTemp2, maxTemp ] = getMaxIndex( mCDBDCT(1:nDCTAtoms,:)' );   
    if maxTemp*SQRT2 > maxCn
        
        maxCn = maxTemp*SQRT2;
        iD1 = 2;
        iD2 = 1;
        iV1 = iTemp1;
        iV2 = iTemp2;
        
    end
        
    [ iTemp1, iTemp2, maxTemp ] = getMaxIndex( mResidule );    
    if maxTemp > maxCn
        
        maxCn = maxTemp;
        iD1 = 2;
        iD2 = 2;
        iV1 = iTemp1;
        iV2 = iTemp2;
        
    end
    
    
    
end


        