function [mImageGrey, bitDepth] = MImage2Grey(mImage,sInfo,map)
% MImage2Grey Converts indexed and truecolor image matrices to a matrix where each
% element represents a greyscale pixel intensity value.


%
% Returns the bitDepth of the image if avaliable.  If not approximtes it
% from the range of pixel intensities.
%
% Usage:    [ mImageGrey, bitDepth ] = MImage2Grey( mImage, sInfo, map )
% Inputs:
%   mImage  matrix representing either a greylevel, truecolor or indexed
%           image.
%   sInfo   meta information from image file
%   map     map of index's to intensities for indexed images
%
% Outputs:
%   mImageGrey  matrix representation of a greyscale image.  Each element
%               of the matrix represents a pixel's intensity.
%   
%   bitDepth    Number of bits required to represent each pixel intensity.
%

functionName = mfilename;
%% Input Validation
if nargin < 2 || nargin > 3
    
    error('%s: Wrong number of input arguments',functionName);
    
elseif nargin == 2
    
    map = [];
    
end
        
% Check ColorType field exists
if isfield(sInfo,'ColorType')
    
    colorType = sInfo.ColorType;
    
else
    
    error('%s: Cannot convert to greyscale as oringinal color type is not availiable!',functionName)
    
end

% Check if bitDepth meta information is avaliable
bitDepth = 0;
if isfield(sInfo,'BitDepth')
        
        bitDepth = sInfo.BitDepth;
        
else
        
        fprintf('%s: BitDepth not avaliable using an approximation!\n',functionName);        
        
end

%% If required convert to grayscale and find the bitDepth of the image
if strcmp(colorType, 'grayscale')
    
    mImageGrey = double(mImage);
    
    if isempty(bitDepth) || bitDepth == 0
        
        bitDepth = ApproxBitdepth(mImageGrey);
        
    end
    
elseif strcmp(colorType, 'truecolor')
        
    mImageGrey = double(rgb2gray(mImage));
    
    if isempty(bitDepth) || bitDepth == 0
        
        bitDepth = ApproxBitdepth(mImageGrey);
        
    else
   
        % Assume 3 colorplanes
        bitDepth = bitDepth/3;
        
    end
    
    fprintf('%s: Supplied image is truecolor, converted to greyscale %i bit image\n',functionName, bitDepth);
    
    
elseif strcmp(colorType, 'indexed')
    
    if isempty(map)
        
        error('%s: Cannot convert indexed image without color map to grey scale!',functionName)
        
    else
        
        mImageGrey = double(ind2gray(mImage,map));
        
        if isempty(bitDepth) || bitDepth == 0
        
            bitDepth= ApproxBitdepth(mImage);
        
        end
        
        fprintf('%s: Supplied image is indexed, converted to greyscale %i bit image\n',functionName, bitDepth);
        
    end

end
%% Validate bitdepth
% As we save the image to the same bitDepth we are given make sure the
% pixel intensities don't exceed this amount
if (2^bitDepth)-1 < max(mImage(:))
    error('%s: %s - bitDepth does not match the range of pixel intensities\n',functionName,sInfo.Filename);
end 