function mRecoveredImage = ExpandImage(foldedImage, cIndex, method, blockWidth, publicKey, privateKey, rotate, mImage, mApproxImage, mEmbeddedImage,sTuning)
% ExpandImage Expands an image from folded image by 1) splitting the host and embedded image 
% and 2) using the embedded coefficients to reconstruct the original.
%
% The procedure is as follows:
% I) Load the image from the imagePath or matrix supplied and split it into \(Q\) blocks \(I_{q}\) of
% blockWidth\(^{2}\) pixels each.
% II) Split the host and embedded image.
% III) Recover the embedded coefficients from the embedded image using the encryption
% operator
% IV) Reconstruct the rest of the original image from embedded coefficients.
%
% Usage:   mRecoveredImage = ExpandImage( foldedImage, cIndex, method, blockWidth, mImage,...
%               publicKey, privateKey, rotate );
%
% Inputs:
% MANDATORY
%   foldedImage     Either the path to where the folded image is saved as a lossless JPEG
%                   file or a matrix representing the folded image.
%   cIndex          A cell structure containing the index's of the atoms in the sparse 
%                   representation.
%   mImageGrey      A matrix representation of the original image converted to grey level
%                   and altered in size so that it can be blocked exactly.
%   method          Approximation method either:
%                       DCT - Discrete Cosine Transform.
%                       RDCTDB - Implementation of OMP in 2D using \(2\) separable 1D
%                       dictionaries (for rows and columns).
%                       The dictionaries used here are \(D_{a}\) and \(D_{b}\) where:
%                           \(D_{a}\): Normalized redundancy \(2\) discrete cosine dictionary.
%                           \(D_{b}\): Dirac basis.
%                       RDCTDBMex - Mex file for implementing the above RDCTDB method.
%   blockWidth      The image is independently processed in \(Q\) square blocks of width
%                   blockWidth. 
%                   Default is \(8\).
%   publicKey       Key used to generate a the random matrix to calculate the vectors
%                   spanning the orthogonal subspace \(S_{K_{q}}^{\bot}\) (c.f. eq. 11 in [1]). This
%                   can be changed publicly to prevent plain text attacks.
%                   Default is \(7\).
%   privateKey      Key used to prevent unauthorized unfolding of the image (c.f. eq. 13 in 
%                   [1]). 
%                   Default \(123456789\).
%   rotate          Determines the method of preventing unauthorized recovery:
%                       \(1\) - the orthogonal basis is randomly rotated (c.f. eq. 13 in [1]).
%                       \(0\) - the orthogonal basis is randomly permuted.
%                   Default is to rotate (\(1\)).
%
% Outputs:
%   mRecoveredImage     A matrix containing the recovered image.

functionName = mfilename;
PrintWithStars([functionName ,' Routine Called']);
%% Input validation
% Check mandatory inputs
if nargin < 4
    fprintf('Path to image cofficient indices, method and blockWidth are mandatory\n');
    return;
end

% Now check optional arguments
if nargin < 5
    publicKey = 123456789;
end
if nargin < 6
    privateKey = 987654321;
end
if nargin < 7
    rotate = 1;
end
compareOriginal = 0;
compareApproximation = 0;
compareEmbedded = 0;
if nargin > 7
    compareOriginal = 1;
end
if nargin > 8
    compareApproximation = 1;
end
if nargin > 9
    compareEmbedded = 1;
end  
if nargin < 11
    sTuning.vTuningFactor = 1;
    sTuning.quantMethod = 'maxQuant';
    sTuning.scaleEmbeddedImage = 0;
    sTuning.displayAnalysis = 0;
end
%% Set up the key structure
sKey.publicKey = publicKey;
sKey.privateKey = privateKey;
sKey.rotate = rotate;
%% Get the foldeimage
if ischar(foldedImage)
    [mFoldedImage, sInfo, map] = LoadImage(foldedImage);
else
    mFoldedImage = foldedImage;
end
%% Get the maxIntensity for PSNR calculations
% As the image is always scaled to the range of intensities we can get the
% maxIntensity as simply the max of the foldedImage
maxIntensity = double(round(max(mFoldedImage(:))));
%% Rescale and validate folded image
M = cIndex{1,1}.M;
m = cIndex{1,1}.m;
mFoldedImage = Rescale(double(mFoldedImage),m,M);
%% Rescale the embeddedImage
if sTuning.scaleEmbeddedImage
    alpham = cIndex{1,1}.alpham;
    alpha = cIndex{1,1}.alpha;
    beta = cIndex{1,1}.beta;
    mFoldedImage = mFoldedImage + alpham - beta;
end
%% Block the image
% nXBlocks will be the same as in the original image nYBlocks would be determined 
% from the number of block separators in the bit stream.
[cFoldedImage, mFoldedImage , nYBlocksFolded, nXBlocks] = BlockImage(mFoldedImage, blockWidth);
nPixelsBlock = blockWidth^2;
%% Calculate size parameters from the cIndex which is calculated from the bitstream
% Here the structure cIndex represents the information we get from the
% bitstream.  This simplifies the situation so that when a bitstream is
% supplied we can just populate cIndex without having to change the rest of
% the code.
nCoefficients = GetNCn(cIndex);
nBlocks = numel(cIndex);
nBlocksUsed = ceil(nCoefficients/nPixelsBlock);
lFree = nBlocksUsed*nPixelsBlock - nCoefficients;
iXBlockEnd = nBlocksUsed - ((nYBlocksFolded-1)*nXBlocks);
nYBlocks = nBlocks/nXBlocks;
if iXBlockEnd == nXBlocks
    iXBlockStart = 1;
    iYBlockStart = nYBlocksFolded+1;
else
    iXBlockStart = iXBlockEnd+1;
    iYBlockStart = nYBlocksFolded;
end
cRecoveredImage = cell(nYBlocks,nXBlocks);
%% Seperate the host and hidden images
% Set up the dictionary here to avoid calculating it for every block
start = clock;
[mDictionary, vSize, mDictionary2D] = Dictionary(method,blockWidth);
[ cRecoveredImage, cEmbeddedRecovery, vRecoveredCn] = ...
    SeparateCellImage(cFoldedImage, cRecoveredImage, mDictionary2D, blockWidth, iXBlockEnd, cIndex, sKey);
% Rescale the embedded image
if sTuning.scaleEmbeddedImage
    mEmbeddedImageRecovered = cell2mat(cEmbeddedRecovery)/alpha;
    cEmbeddedRecovery = BlockImage(mEmbeddedImageRecovered, blockWidth);
    vRecoveredCn = GetCnFromEmbeddedImage(cEmbeddedRecovery, mDictionary2D, blockWidth, iXBlockEnd, cIndex, sKey);
end
removeTime = etime(clock,start);
fprintf('Time to seperate the embedded image (F) from the host image: %.2f seconds\n',removeTime);
%{
Remove lFree padding from  vHiddenCoefficients.  If we were not
using an orthogonal basis to hide the coefficients then we would
distribute this space over all the blocks to avoid the calculation of the
projector being ill posed, as the proctor will be orthogonal this step is
unecessary.
%}
vRecoveredCn = vRecoveredCn(1:end-lFree);
%% Reverse the quantisation
if ~strcmp(sTuning.quantMethod,'none')
    cQuant = cIndex{1,1}.cQuant;
    % Get the original index's of all the hidden coefficients
    vHiddenIndex = GetFoldedCnIndex(cIndex, blockWidth, iYBlockStart, iXBlockStart);
    % Reverse the quantisation
    vRecoveredCn = RRescaleCoefficients(vRecoveredCn,vHiddenIndex,cQuant, sTuning.displayAnalysis);
end
%% Reconstruct the rest of the image using vHiddenCoefficients
start = clock;
cRecoveredImage = ReconstructCellImage(cRecoveredImage, vRecoveredCn,...
    mDictionary2D, cIndex, blockWidth, iYBlockStart, iXBlockStart);
reconstructTime = etime(clock,start);
fprintf('Time to reconstruct the rest of the image: %.2f seconds\n',reconstructTime);
mRecoveredImage = cell2mat(cRecoveredImage);
% if original image is provided compare with hidden to get PSNR
mHostImageRecovered = GetSubImage(cell2mat(cRecoveredImage),[1,1], [nYBlocksFolded, iXBlockEnd], blockWidth);
mHiddenImageRecovered = GetSubImage(cell2mat(cRecoveredImage),[iYBlockStart, iXBlockStart],[nYBlocks,nXBlocks], blockWidth);
if sTuning.displayAnalysis
    %% Display the results
    if compareApproximation
        mHostImage = GetSubImage(mApproxImage,[1,1], [nYBlocksFolded, iXBlockEnd], blockWidth);
        mHiddenImage = GetSubImage(mApproxImage,[iYBlockStart, iXBlockStart],[nYBlocks,nXBlocks], blockWidth);
        PrintWithStars('PSNR between approximation and recovery splits:');
        PSNRHost = CalcPSNR(mHostImage,mHostImageRecovered, maxIntensity);
        fprintf('Host image: %.4f dB\n',PSNRHost);
        PSNRHidden = CalcPSNR(mHiddenImage,mHiddenImageRecovered, maxIntensity);
        fprintf('Hidden image: %.4f dB\n',PSNRHidden);
    end
    if compareApproximation
        PSNRHostPlusHidden = CalcPSNR(mApproxImage,mRecoveredImage, maxIntensity);
        fprintf('Whole Image: %.4f dB\n',PSNRHostPlusHidden);
    end
    if compareEmbedded
        for i = iXBlockEnd+1:nXBlocks
            cEmbeddedRecovery{end,i} = zeros(blockWidth);
        end
        mEmbeddedImageRecovered = cell2mat(cEmbeddedRecovery);
        PSNREmbedded = CalcPSNR(mEmbeddedImage,mEmbeddedImageRecovered, maxIntensity);
        PrintWithStars('PSNR between F the embedded image (c.f eq. 9 in [1]) we add and the embedded image we recover:');
        fprintf('%.4f dB\n',PSNREmbedded);
    end

end
if compareOriginal
    PrintWithStars('PSNR between the Original image and the recovery');
    PSNROriginalRecovered = CalcPSNR(round(mImage),round(mRecoveredImage), maxIntensity);
    fprintf('%.4f dB\n',PSNROriginalRecovered);
end
end

%% Begin sub functions

function cRecoveredImage = ReconstructCellImage(cRecoveredImage, vCn,...
    mDictionary, cIndex, blockWidth, iYBlockStart, iXBlockStart)

    [nYBlocks, nXBlocks] = size(cRecoveredImage);
    rowIncrement = 1;
    iCn  = 1;
    for i = iYBlockStart:nYBlocks

        for j = iXBlockStart:rowIncrement:nXBlocks

            % Construct the matix of atoms used to reconstruct this block
            iMDictionary = cIndex{i,j}.iMDictionary2D;
            mImageSpace = mDictionary(:,iMDictionary);
            % Get the coefficients to recover this block
            nCn = size(iMDictionary,2);
            vCnBlock = vCn(iCn:iCn+nCn-1);
            iCn = iCn + nCn;

            % Recover the block
            cRecoveredImage{i,j} = zeros(blockWidth);
            cRecoveredImage{i,j}(:) = mImageSpace*vCnBlock;

        end
        if i == iYBlockStart
            iXBlockStart = 1;
        end
        temp = iXBlockStart;
        iXBlockStart = nXBlocks;
        nXBlocks = temp;
        rowIncrement = rowIncrement * -1;

    end

end

function nCn = GetNCn(cIndex)

[nYBlocks, nXBlocks] = size(cIndex);
nCn = 0;
for i = 1:nYBlocks
    for j = 1:nXBlocks
        nCn = nCn + numel(cIndex{i,j}.iMDictionary2D);
    end
end

end

function vHiddenIndex = GetFoldedCnIndex(cIndex, blockWidth, iYBlockStart, iXBlockStart)

    [nYBlocks, nXBlocks] = size(cIndex);
    % Allocate memory for variables
    vHiddenIndex = zeros(blockWidth^2*nYBlocks*nXBlocks,1);
    rowIncrement = 1;
    iVHiddenIndex  = 1;
    for i = iYBlockStart:nYBlocks

        for j = iXBlockStart:rowIncrement:nXBlocks

            iMDictionary2D = cIndex{i,j}.iMDictionary2D;
            nCn = size(iMDictionary2D,2);
            vHiddenIndex(iVHiddenIndex:iVHiddenIndex + nCn -1) = iMDictionary2D;
            iVHiddenIndex = iVHiddenIndex + nCn;     

        end
        if i == iYBlockStart
            iXBlockStart = 1;
        end
        temp = iXBlockStart;
        iXBlockStart = nXBlocks;
        nXBlocks = temp;
        rowIncrement = rowIncrement * -1;

    end

end