function [ foldedImage, cIndex, mImageGrey, maxIntensity, mApproxImage, mEmbeddedImage] = FoldImage(imagePath, approxMethod, PSNR, blockWidth, publicKey, privateKey, rotate, saveImage, sTuning)
% FoldImage Loads an image and folds it using a public and private key.
%
% When viewing an image from top to bottom this routine folds the bottom section into a
% sparse representation of the top section.  The procedure is as follows:
% I) Load the image from the supplied imagePath and split it into \(Q\) blocks \(I_{q}\) containing 
% \(N_{q} \mbox{x} N_{q} = \) blockWidth\(^2\) pixels each.
% II) Find a sparse representation of each block \(I_{q}^{K_{q}}\) in a subspace \(S_{K_{q}}\) (c.f. eq. 10 in 
% [1]).
% III) The number of host blocks \(H\) depends on the sparsity in the approximation of the
% whole image \(I\). For each of these \(H\) blocks we calculate a basis for \(S_{K_{q}}^{\bot}\) (the orthogonal 
% complement of \(S_{K_{q}}\) in  \(R^{N_{q}} \mbox{x} R^{N_{q}}\)).
% IV) We map the coefficients of the sparse representation of the remaining \(Q - H\) blocks
% onto the subspace \(S_{K_{q}}^{\bot}\) constructing in that way the matrices \(F_{q}\) in \(S_{K_{q}}^{\bot}\) (c.f. eq. 8 in [1]).
% V) We add each matrix \(F_{q}\) to the corresponding host block \(I_{q}^{K_{q}}\) to obtain \(G_{q} = F_{q} + I_{q}^{K_{q}}\), 
% with \(I_{q}^{K_{q}}\) in \(S_{K_{q}}\) and \(F_{q}\) in \(S_{K_{q}}^{\bot}\).
%
% Usage:        [ foldedImage, cIndex, mImageGrey, maxIntensity] = FoldImage( imagePath,...
%                   approxMethod, approxCriteria, blockWidth, publicKey, privateKey,...
%                   rotate, saveImage );
%
% Inputs:
% MANDATORY
%   imagePath   Relative or absolute path to the location of the image file including the
%               image file name.
% OPTIONAL
%   method      Approximation method either:
%                   DCT - Discrete Cosine Transform.
%                   RDCTDB - Implementation of OMP in 2D using \(2\) separable 1D dictionaries 
%                   (for rows and columns).
%                   The dictionaries used here are \(D_{a}\) and \(D_{b}\) where:
%                       \(D_{a}\): Normalized redundancy \(2\) discrete cosine dictionary.
%                       \(D_{b}\): Dirac basis.
%                   RDCTDBMex - Mex file for implementing the above RDCTDB method.
%               Default is RDCTDB.
%   PSNR        Desired PSNR of image approximation.  The result will be a greater PSNR for
%               the whole image as each block must individually satisfy this criteria.  
%               Default is \(40\)dB.
%   blockWidth  The image is independently processed in \(Q\) square blocks of width blockWidth. 
%               Default is \(8\).
%   publicKey   Key used to generate a the random matrix to calculate the vectors spanning 
%               the orthogonal subspace \(S_{K_{q}}^{\bot}\) (c.f. eq. 11 in [1]). This can be changed 
%               publicly to prevent plain text attacks.
%               Default is \(7\).
%   privateKey  Key used to prevent unauthorized unfolding of the image (c.f. eq. 13 in [1]). 
%               Default \(123456789\).
%   rotate      Determines the method of preventing unauthorized recovery:
%                   \(1\) - the orthogonal basis is randomly rotated (c.f. eq. 13 in [1]).
%                   \(0\) - the orthogonal basis is randomly permuted.
%               Default is to rotate (\(1\)).
%   saveImage   Determines what is returned to the variable foldedImage.
%                   \(1\) - path to where the folded image is saved as a lossless JPEG.
%                   \(0\) - matrix representation of the folded image.
%               Default is to save (\(1\)).
%
% Outputs:
%   foldedImage     Either the path to where the folded image is saved as a lossless JPEG
%                   file or a matrix representing the folded image.
%   cIndex          A cell structure containing the index's of the atoms in the sparse 
%                   representation.
%   mImageGrey      A matrix representation of the original image converted to grey level
%                   and altered in size so it can be blocked exactly.
%   maxIntensity    Maximum size of pixel intensity that can be used.

functionName = mfilename;
PrintWithStars([functionName ,' Routine Called']);
%% Input validation
% Check mandatory inputs
if nargin < 1
    fprintf('%s: Path to image not supplied!\n',functionName);
    return;
end

% Now check optional arguments
if nargin < 2
    method = 'RDCTDB';
end
if nargin < 3 
    PSNR = 40;
end
if nargin < 4
    blockWidth = 8;
end
if nargin < 5
    publicKey = 7;
end
if nargin < 6
    privateKey = 123456789;
end
if nargin < 7
    rotate = 1;
end
if nargin < 8
    saveImage = 0;
end
if nargin < 9
    sTuning.vTuningFactor = 1;
    sTuning.quantMethod = 'maxQuant';
    sTuning.scaleEmbeddedImage = 0;
    sTuning.displayImages = 0;
    sTuning.displayAnalysis = 0;
end
%% Set up the key structure
sKey.publicKey = publicKey;
sKey.privateKey = privateKey;
sKey.rotate = rotate;
%% Load, validate and block the image
[mImage, sInfo, map] = LoadImage(imagePath);

% Convert to grey scale
[mImageGrey, bitDepth] = MImage2Grey(mImage, sInfo, map);
[y x] = size(mImageGrey);
maxIntensity = (2^bitDepth) -1;

% As we are saving as JPEG make sure the bitDepth doesn't exceed 16
if bitDepth > 16
    error('%s: Bit depth: %i exceeds bit depth that we can save file',functionName,bitDepth);
end

% Block the image
[cImageGrey, mImageGrey, nYBlocks, nXBlocks, y, x] = BlockImage(mImageGrey, blockWidth);
%% Get sparse representation
% Set up the dictionary here to avoid calculating it for every block
[mDictionary,vSize,mDictionary2D] = Dictionary(approxMethod,blockWidth);
start = clock;
blockError = PSNRToBlockErrorNorm( PSNR, blockWidth^2, maxIntensity );
if strcmp(approxMethod, 'DCT')
    
    [ cImageApprox, nCoefficients, cCoefficients, cIndex, nZeroBlocks ] = ProcessCellImageDCT(cImageGrey,blockError);
    
elseif strcmp(approxMethod, 'RDCTDB')
    
    [ cImageApprox, nCoefficients, cCoefficients, cIndex, nZeroBlocks ] = ProcessCellImageRDCTDB(cImageGrey, blockError);
    
elseif strcmp(approxMethod, 'RDCTDBMex')
    
    [ cImageApprox, nCoefficients, cCoefficients, cIndex, nZeroBlocks ] = ...
        ProcessCellImageGreedy2D(cImageGrey, blockError, mDictionary, vSize, 'OMP2D');
    
else
    
    error('%s: Approximation Method Unknown',functionName);
    
end
elapsedTime = etime(clock,start);
%% Calculate approximation quality and folded image dimensions
nPixels = y*x;
PSNR = CalcPSNR(mImageGrey,round(cell2mat(cImageApprox)),maxIntensity);
SR = nPixels/nCoefficients;
PrintWithStars(['Approximation Processing ', imagePath, ' in blocks of ' ...
    num2str(blockWidth), 'x', num2str(blockWidth), ' pixels using ' approxMethod]);
fprintf('Approximation time: %.2g seconds\n',elapsedTime);
% Print message to encourage use of mex file.
if strcmp(approxMethod, 'RDCTDB')
    fprintf('Approximaton time can be reduced by up to 10 times by using the RDCDBMex option!\n');
end
fprintf('PSNR in approximation: %.4f dB for an SR of %.4f\n',PSNR,SR);
if nZeroBlocks > 0
    fprintf('Number of zero energy blocks: %i.\n',nZeroBlocks);
end

 %{ 
Calculate the dimensions of the host and folded image as nYBlocksNew X nXBlocks where
iXBlockEnd is the index of the last block used on row nYBlocksNew.  The remaining
nXBlocks - iXBlocksEnd are padded with zeros.
lFree is the amount of free positions left - re-write?????
%}
[nYBlocksNew, iYBlockStart, iXBlockEnd, iXBlockStart, lFree, AdjustedSR] = ...
    CalculateNewDimensions(nPixels, nCoefficients, nXBlocks, blockWidth);
if sTuning.displayAnalysis
    PrintWithStars('Dimensions');
    fprintf('Original Image: %ix%i pixels, %ix%i blocks\n',y,x,nYBlocks,nXBlocks);
    fprintf('Folded Image: %ix%i pixels, %ix%i blocks\n',nYBlocksNew*blockWidth,x,nYBlocksNew,nXBlocks);
    fprintf('Row index of last block: %i, Space left: %i\n',iXBlockEnd,lFree);
    fprintf('Adjusted SR: %.4f\n',AdjustedSR);
end
%% Get the coefficients we will fold into the host image
% Place the coeffients not used for the host image in a vector vCnHidden
[vCnHidden,vHiddenIndex] = GetCnToFold(cCoefficients,cIndex,iXBlockStart,iYBlockStart,nYBlocks, nXBlocks, nCoefficients);
%{
Pad vHiddenCoefficients with zeros to accomodate lFree.  If we were not
using an orthogonal basis to hide the coefficients then we would
distribute this space over all the blocks to avoid the calculation of the
projector being ill posed, as the proctor will be orthogonal this step is
unecessary.
%}
vCnHidden = [vCnHidden; zeros(lFree,1)];
%% Quantize the hidden coefficients if required
if ~strcmp(sTuning.quantMethod,'none')
    % Get the index's of the coefficients to quantize
    vQuantIndex = GetQuantIndex(approxMethod, blockWidth);
    % Get the quantization factors
    cQuant = GetQuantFactor(vCnHidden,vHiddenIndex,vQuantIndex,sTuning.quantMethod, sTuning.vTuningFactor,sTuning.displayAnalysis);
    % Quantize the hidden coefficients and return as a vector
    vCnHidden = RescaleCoefficients(vCnHidden,vHiddenIndex,cQuant,sTuning.displayAnalysis);
    % Store the quant factors in the bitstream
    cIndex{1,1}.cQuant = cQuant;
end
%% Begin Folding
% Calculate the embedded image
start = clock;
mEmbeddedImage = GenerateEmbeddedImage(vCnHidden, mDictionary2D, cIndex(1:nYBlocksNew,:), iXBlockEnd, sKey);
elapsedTime = etime(clock,start);
PrintWithStars('Folding');
fprintf('Time: %.2f seconds\n',elapsedTime);
if sTuning.displayAnalysis
    fprintf('Key to change to prevent plain text attacks %i\n',publicKey);
    fprintf('Key to prevent unorthorized recovery %i\n',privateKey);
    if rotate
        fprintf('Orthogonal basis hidden using random rotation\n');
    else
        fprintf('Orthogonal basis hidden using random permutation\n');
    end
    minEmbeddedImage = min(mEmbeddedImage(:));
    maxEmbeddedImage = max(mEmbeddedImage(:));
    meanEmbeddedImage = mean(mEmbeddedImage(:));
    fprintf('Range of intensities of embedded image\n');
    fprintf('Min: %.0f, Max %.0f, Mean: %.2f\n',minEmbeddedImage,maxEmbeddedImage,meanEmbeddedImage);
end
%% Rescale the embedded image to the same range as the image
% Remember to ignore zeros refine this needs to return min and max for
% actual host here we are rescaling including zero padding
if sTuning.scaleEmbeddedImage
    mHostBig = cell2mat(cImageApprox(1:nYBlocksNew,:));
    imageMin = min(mHostBig(:));
    imageMax = max(mHostBig(:));
    [mEmbeddedImage m M] = Rescale(mEmbeddedImage,imageMin,imageMax);
    cIndex{1,1}.mEmbeddedImage = M;
    cIndex{1,1}.mEmbeddedImage = m;
    cIndex{1,1}.beta = imageMin;
    cIndex{1,1}.alpha = (imageMax - imageMin)/(M-m);
    cIndex{1,1}.alpham = m*cIndex{1,1}.alpha;
end

%% Place the images into matrices
mApproxImage = cell2mat(cImageApprox);
cHostImage = cImageApprox(1:nYBlocksNew,:);
clear cImageApprox;
for i = iXBlockEnd+1:nXBlocks
    cHostImage{end,i} = zeros(blockWidth);
end
mHostImage = cell2mat(cHostImage);
mFoldedImage = mHostImage + mEmbeddedImage;

%% Rescale
[mFoldedImageRescaled m M] = Rescale(mFoldedImage,0,maxIntensity);
cIndex{1,1}.M = M;
cIndex{1,1}.m = m;
scalingFactor = (M-m)/maxIntensity;
if saveImage
    %%  Save
    iLastForwardSlash = max(strfind(imagePath,'/'));
    % If we have been given a file path instead of just the file name trim off
    % the path
    if iLastForwardSlash
        fileName = imagePath(iLastForwardSlash+1:end);
    else
        fileName = imagePath;
    end
    % Remove file extension
    iDot = max(strfind(fileName,'.'));
    fileName = fileName(1:iDot-1); 
    
    % Assume FoldImage.m is in the path EIF/Folding_Routines/, we want to save the image to
    % EIF/Images/folded_Images relative to the location of FoldImage.   
    thisFunctionPath = which(functionName);    
    iFileSep = strfind(thisFunctionPath,filesep);
    saveImagePath = [thisFunctionPath(1:iFileSep(end-1)), 'Images/Folded_Images/'];
    % If folder doesn't exist then create it
    if exist(saveImagePath) ~= 7
        mkdir(saveImagePath);
    end
    foldedImage = [saveImagePath 'folded_' fileName '.jpeg'];
    % Convert image to the correct format
    if bitDepth <= 8
        mFoldedImageRescaled = uint8(mFoldedImageRescaled);
        bitDepthToWrite = 8;
    elseif bitDepth <= 12
        mFoldedImageRescaled = uint16(mFoldedImageRescaled);
        bitDepthToWrite = 12;
    elseif bitDepth <= 16        
        mFoldedImageRescaled = uint16(mFoldedImageRescaled);
        bitDepthToWrite = 16;
    else
        error('%s: Bit depth not compatible with JPEG file format',functionName);
    end
    % Save as lossless jpeg
    imwrite(mFoldedImageRescaled,foldedImage,'Bitdepth',bitDepthToWrite,'Mode','lossless');

    if sTuning.displayAnalysis
        %% If saving then print out the loss from the rescaling operation
        PrintWithStars('Rescaled');
        fprintf('Rescaling Factor: %.2f\n',scalingFactor);
        rescaledLoss = CalcPSNR(mFoldedImage,Rescale(round(mFoldedImageRescaled),m,M));
        fprintf('Loss of %.4f dB in G = H + F from rescaling and saving to %i Bits\n',rescaledLoss,bitDepth);
        fprintf('Image saved as %s\n',foldedImage');
    end
    fprintf('Image saved as %s\n',foldedImage');
else
    foldedImage = mFoldedImageRescaled;
end

if sTuning.displayImages
    %% Display Images
    scrsz = get(0,'ScreenSize');
    figure('OuterPosition',[1 1 scrsz(3)/2 scrsz(4)]);
    subplot(2,2,1)
    imshow(mImageGrey,[0,maxIntensity]);
    title(['Original Image (I): ',fileName]);
    subplot(2,2,2)
    imshow(mHostImage,[0,maxIntensity]);
    title(['Host Image ', approxMethod]);
    subplot(2,2,3)
    imshow(mEmbeddedImage,[0,maxIntensity]);
    title('Embedded Image (F) (c.f. eq. 8 in [1])');
    subplot(2,2,4)
    imshow(mFoldedImage,[0,maxIntensity]);
    title('Folded Image (G = Host + F) ');
end

end


%% Start of sub functions
function [nYBlocks, iYBlockStart, iXBlockEnd, iXBlockStart, lFree, AdjustedSR] = ...
    CalculateNewDimensions(nPixels, nCoefficients, nXBlocks, blockWidth)

    nBlocks = ceil(nCoefficients/blockWidth^2);
    lFree = (nBlocks*blockWidth^2) - nCoefficients;
    nYBlocks = ceil(nBlocks/nXBlocks);
    iXBlockEnd = nXBlocks - (nYBlocks*nXBlocks - nBlocks);
    if iXBlockEnd == nXBlocks
        iXBlockStart = 1;
        iYBlockStart = nYBlocks +1;
    else    
        iXBlockStart = iXBlockEnd  + 1;
        iYBlockStart = nYBlocks;
    end
    AdjustedSR = nPixels/(nCoefficients+lFree);

end

function [vHiddenCn,vHiddenIndex] = GetCnToFold(cCn,cIndex,...
    iXBlockStart,nYBlocksNew,nYBlocks, nXBlocks, nCoefficients)

    vHiddenCn = zeros(nCoefficients,1);
    vHiddenIndex = zeros(nCoefficients,1);
    iHidden = 1;
    rowIncrement = 1;
    %% Scan in the order we will apply the DPCM
    for i = nYBlocksNew:nYBlocks

        for j = iXBlockStart:rowIncrement:nXBlocks

            nCn = numel(cCn{i,j});        
            vHiddenCn(iHidden:iHidden+nCn-1) = cCn{i,j};
            vHiddenIndex(iHidden:iHidden+nCn-1) = cIndex{i,j}.iMDictionary2D;
            iHidden = iHidden + nCn;

        end
        if i == nYBlocksNew
            iXBlockStart = 1;
        end
        temp = iXBlockStart;
        iXBlockStart = nXBlocks;
        nXBlocks = temp;
        rowIncrement = rowIncrement * -1;
    end
    vHiddenCn = vHiddenCn(1:iHidden-1);  

end

function vQuantIndex = GetQuantIndex(approxMethod,blockWidth)

    % In future will be adaptive based on coefficient values.
    if strcmp(approxMethod, 'DCT')

            vQuantIndex = 1;

    elseif strcmp(approxMethod(1:6), 'RDCTDB') 

        iD1 = [1;1;1;1];
        iV1 = [1;1;2;2];
        iD2 = [1;1;1;1];
        iV2 = [1;2;1;2];
        vDictSize = [(2*blockWidth-1); blockWidth];
        vQuantIndex = KroneckerIndex(iD1,iD2,iV1,iV2,vDictSize);

    end
    
end