function cQuant = GetQuantFactor(vCnHidden,vHiddenIndex,vQuantIndex,method, tuningFactor, displayAnalysis)

% Validate method
if nargin < 5
    tuningFactor = 1;
end
if nargin < 6
    displayAnalysis = 0;
end
if ~(strcmp(method,'minQuant') || strcmp(method,'maxQuant') || strcmp(method,'rescale') || strcmp(method,'none'))
    
    error('Quantization method unknown!');
    
end
cQuant.method = method;
if ~strcmp(method,'none')
    %% Get the the index and values of the coefficients we want to quantize
    nQuant = numel(vQuantIndex);
    cVCnToQuant = cell(nQuant,1);
    vNCnToQuant  = zeros(nQuant,1); 
    nVHiddenCn = numel(vCnHidden);
    vICnToQuant = zeros(nVHiddenCn,1);
    iVICnToQuant = 1;
    vQuantIndexFound = zeros(nQuant,1);
    iVQuantIndexFound = 0;
    for n =1:nQuant       
        % Store the coefficients we want to quantize in a cell of size
        % nQuant x 1 where  cVCnToQuant{n} contains a vector of coefficinets
        % corresponding the index vHiddenIndex(n)
        vICnToQuantLoop = find(vHiddenIndex == vQuantIndex(n));
        vNCnToQuant(n) = numel(vICnToQuantLoop);
        if vNCnToQuant(n) > 0
            iVQuantIndexFound = iVQuantIndexFound + 1;
            vQuantIndexFound(iVQuantIndexFound) = vQuantIndex(n);
            cVCnToQuant{iVQuantIndexFound} = vCnHidden(vICnToQuantLoop);

            % Store all the indexes in a vector
            vICnToQuant(iVICnToQuant:iVICnToQuant+vNCnToQuant(n)-1) = vICnToQuantLoop;
            iVICnToQuant =  iVICnToQuant + vNCnToQuant(n);
        end

    end
    vICnToQuant = vICnToQuant(1:iVICnToQuant-1);
    vQuantIndexFound  = vQuantIndexFound(1:iVQuantIndexFound);
else
    vICnToQuant = [];
end
%% Store the min max and average value for all coefficints
vCnHiddenSorted = sort(abs(vCnHidden)); % Ignore zero coefficients
minCnHidden = vCnHiddenSorted(min(find(vCnHiddenSorted > 0)));
clear vCnHiddenSorted;
maxCnHidden = max(abs(vCnHidden));
meanCnHidden = mean(abs(vCnHidden));
%% Now deal with the coefficients we don't quantizize
% Performed here so we can report the ranges of coefficinets even if no
% quantization takes place.
%% Get the indexes of these coefficients
vICnNoQuant   = setdiff(1:nVHiddenCn,vICnToQuant);
vCnNoQuant = vCnHidden(vICnNoQuant); 
%% Store the min max and average value for coefficients at each index we don't quantize
vNoQuantSorted = sort(abs(vCnNoQuant)); % Ignore zero coefficients
minCnNoQuant = vNoQuantSorted(min(find(vNoQuantSorted > 0)));
clear vNoQuantSorted;
maxCnNoQuant = max(abs(vCnNoQuant));
meanCnNoQuant = mean(abs(vCnNoQuant));
% To avoid dividing by zero check that all remaining coefficinets are not zero.  
% As minCnNoQuant is the smallest non zero element we can validate the
% vector based on this.
if minCnNoQuant == 0
    error('All coefficinets we dont quantize are zero!');
end
%% Deal with the coefficients we want to quantize
if iVQuantIndexFound > 0    
    %% Store the min max and average value for these coefficients
    vMinCnToQuant = zeros(iVQuantIndexFound,1);
    vMaxCnToQuant = zeros(iVQuantIndexFound,1);
    vMeanCnToQuant = zeros(iVQuantIndexFound,1);
    for n = 1:iVQuantIndexFound
        vQuantSorted = sort(abs(cVCnToQuant{n})); % Ignore zero coefficients
        vMinCnToQuant(n) = vQuantSorted(min(find(vQuantSorted > 0)));
        clear vQuantSorted;
        vMaxCnToQuant(n) = max(abs(cVCnToQuant{n})); 
        vMeanCnToQuant(n) = mean(abs(cVCnToQuant{n}));
    end

    %% Calculate quntization factors
    vMinCnNew = zeros(iVQuantIndexFound,1);
    vMaxCnNew = zeros(iVQuantIndexFound,1);
    vMeanCnNew = zeros(iVQuantIndexFound,1);
    if strcmp(method,'minQuant')
        vMinCnNew = ones(iVQuantIndexFound,1)*minCnNoQuant*tuningFactor;
        quantFactors = vMinCnToQuant./vMinCnNew;
        vMaxCnNew = vMaxCnToQuant./quantFactors;
        vMeanCnNew = vMeanCnToQuant./quantFactors;
        rQuantFactors = 1./quantFactors;
    elseif strcmp(method,'maxQuant')
        vMaxCnNew = ones(iVQuantIndexFound,1)*maxCnNoQuant*tuningFactor;
        quantFactors = vMaxCnToQuant./vMaxCnNew;
        vMinCnNew = vMinCnToQuant./quantFactors;
        vMeanCnNew = vMeanCnToQuant./quantFactors;       
        rQuantFactors = 1./quantFactors;
    elseif strcmp(method,'rescale')
        vMinCnNew = ones(iVQuantIndexFound,1)*minCnNoQuant*tuningFactor(1);
        vMaxCnNew = ones(iVQuantIndexFound,1)*maxCnNoQuant*tuningFactor(2);
        vMeanCnNew = (vMaxCnNew-vMinCnNew).* (vMeanCnToQuant-vMinCnToQuant)./(vMaxCnToQuant-vMinCnToQuant) + vMinCnNew;
        quantFactors(:,1) = vMinCnNew;
        quantFactors(:,2) = vMaxCnNew;
        rQuantFactors(:,1) = vMinCnToQuant;
        rQuantFactors(:,2) = vMaxCnToQuant;        
    end
        
    %% Populate the cQuant structure
    cQuant.vQuantIndex = vQuantIndexFound;
    cQuant.quantFactors = quantFactors;
    cQuant.rQuantFactors = rQuantFactors;  
    
else
    
    fprintf('No Quantization as no coefficients found at specified indexes!\n');
    cQuant.method = 'none';
    
end
            
if displayAnalysis
    %% Print out effects of quantization
    PrintWithStars('Planned Quantization');
    fprintf('All Existing Hidden Coefficients\n');
    fprintf('Min: %.2f, Max %.2f, Mean: %.2f\n',minCnHidden,maxCnHidden,meanCnHidden);
    fprintf('Remaining Existing Hidden Coefficients\n');
    fprintf('Min: %.2f, Max %.2f, Mean: %.2f\n',minCnNoQuant,maxCnNoQuant,meanCnNoQuant);
    if iVQuantIndexFound > 0
        fprintf('By Index:');
        for i = 1:iVQuantIndexFound
            fprintf('  Index: %i\n',vQuantIndexFound(i));
            fprintf('    Current :- Min: %.2f, Max %.2f, Mean: %.2f\n',vMinCnToQuant(i),vMaxCnToQuant(i),vMeanCnToQuant(i));
            fprintf('    Planned :- Min: %.2f, Max %.2f, Mean: %.2f\n',vMinCnNew(i),vMaxCnNew(i),vMeanCnNew(i));
        end
    end
end
